from typing import Dict

import collections
import json
import os
import yaml

from copy import deepcopy
from pathlib import Path as Path


def deep_merge(dict1, dict2):
    """ Return a new dictionary by merging two dictionaries recursively. """

    result = deepcopy(dict1)

    for key, value in dict2.items():
        if isinstance(value, collections.Mapping):
            result[key] = deep_merge(result.get(key, {}), value)
        else:
            result[key] = deepcopy(dict2[key])

    return result


def expand_environment_variables(contents: str) -> str:
    return os.path.expandvars(contents)


def load_config(path: Path, expand_envars=True, must_exist=False) -> Dict:
    if path.exists():
        contents = path.read_text()
        if expand_envars:
            contents = expand_environment_variables(contents)

        if path.suffix in (".json",):
            config = json.loads(contents)
        elif path.suffix in (".yml", ".yaml"):
            config = yaml.safe_load(contents)
        else:
            raise TypeError(f"Cannot parse '{path.suffix}' files!")

        return config

    elif must_exist:
        raise IOError(f"{path} does not exist!")

    else:
        return {}



from typing import Iterator, Optional

from eth_account.messages import SignableMessage  # type: ignore
from eth_account.datastructures import SignedMessage, SignedTransaction  # type: ignore


class AccountAPI:
    @property
    def address(self) -> str:
        ...

    def sign_message(self, msg: SignableMessage) -> Optional[SignedMessage]:
        ...

    def sign_transaction(self, txn: dict) -> Optional[SignedTransaction]:
        ...

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.address}>"

    def __str__(self) -> str:
        return self.address


class AccountControllerAPI:
    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[AccountAPI]:
        ...

    def __getitem__(self, address: str) -> AccountAPI:
        for account in self.__iter__():
            if account.address == address:
                return account

        raise IndexError(f"No local account {address}.")

    def __contains__(self, address: str) -> bool:
        try:
            self.__getitem__(address)
            return True
        except IndexError:
            return False



from pathlib import Path as _Path

# Path constants for Ape
INSTALL_FOLDER = _Path(__file__).parent
DATA_FOLDER = _Path.home().joinpath(".ape")
DATA_SUBFOLDERS = ("accounts",)
PROJECT_FOLDER = _Path.cwd()



# NOTE: Modules in Python are singletons, this module implements
#       all of the configuration items for the ape tool
# https://docs.python.org/3/faq/programming.html#how-do-i-share-global-variables-across-modules
from typing import Dict
import sys as _sys

from pathlib import Path

from .constants import (
    INSTALL_FOLDER,
    PROJECT_FOLDER,
    DATA_FOLDER,
)
from ._utils import (
    deep_merge as _deep_merge,
    load_config as _load_config,
)

from ape import __version__


# For all HTTP requests we make
_python_version = (
    f"{_sys.version_info.major}.{_sys.version_info.minor}"
    f".{_sys.version_info.micro} {_sys.version_info.releaselevel}"
)

REQUEST_HEADERS = {
    "User-Agent": f"Ape/{__version__} (Python/{_python_version})",
}


# Our example config file, with all the defaults set
DEFAULT_CONFIG = _load_config(
    INSTALL_FOLDER.joinpath("data/default-config.yaml"),
    must_exist=True,
)

# The config from the user's environment
_user_config = _load_config(PROJECT_FOLDER.joinpath("ape-config.yaml"))


# Configuration items (these can be modified from their initial state)
def _get_merged_config(name: str) -> Dict:
    return _deep_merge(
        DEFAULT_CONFIG[name],
        _user_config.get(name, {}),
    )


def load_config(path: Path) -> Dict:
    return _deep_merge(DEFAULT_CONFIG, _load_config(path))


# Do this so we have typing info for these objects
project_structure = _get_merged_config("project_structure")
compiler = _get_merged_config("compiler")
console = _get_merged_config("console")
hypothesis = _get_merged_config("hypothesis")
autofetch_sources = _get_merged_config("autofetch_sources")
dependencies = _get_merged_config("dependencies")
dev_deployment_artifacts = _get_merged_config("dev_deployment_artifacts")
_user_network_settings = _get_merged_config("networks")

# Allow using custom (unsupported) config items
custom = {k: v for k, v in _user_config.items() if k not in DEFAULT_CONFIG.keys()}

# Set up network configuration (these will change over time)
active_network = _user_network_settings.get("default", "development")

networks = _load_config(DATA_FOLDER.joinpath("network-config.yaml"))

# Merge project config file settings with global network settings
for name, params in networks["local"].items():
    networks["local"][name] = _deep_merge(params, _user_network_settings["local"])

for name, params in networks["public"].items():
    networks["public"][name] = _deep_merge(params, _user_network_settings["public"])


def _get_network_type(id: str) -> str:
    if id not in (*networks["local"].keys(), *networks["public"].keys()):
        raise ValueError(f"'{id}' is not a valid network name")

    if id in networks["local"].keys():
        return "local"
    else:
        return "public"


def get_active_network_params() -> Dict:
    if active_network is None:
        raise ConnectionError("No active network")

    typ = _get_network_type(active_network)
    if "fork" in active_network:
        # Handle "fork mode" differently
        fork = active_network.replace("-fork", "")
        params = _deep_merge(
            networks[_get_network_type(fork)][fork],
            networks[typ][active_network],
        )
        # This is the key difference
        params["cmd_settings"]["fork"] = networks[typ][fork]["host"]
        return params
    else:
        return networks[typ][active_network]


def set_active_network(id: str):
    _get_network_type(id)  # Raises if not valid
    global active_network
    active_network = id



from importlib.metadata import version, PackageNotFoundError

from . import _setup  # NOTE: Do this before anything else

try:
    __version__ = version("ape")
except PackageNotFoundError:
    # package is not installed
    pass



# NOTE: This file only executes once on module load
import shutil
from pathlib import Path

from .constants import (
    INSTALL_FOLDER,
    DATA_FOLDER,
    DATA_SUBFOLDERS,
)


# create data folder structure
DATA_FOLDER.mkdir(exist_ok=True)
for folder in DATA_SUBFOLDERS:
    DATA_FOLDER.joinpath(folder).mkdir(exist_ok=True)

if not DATA_FOLDER.joinpath("network-config.yaml").exists():
    shutil.copyfile(
        INSTALL_FOLDER.joinpath("data/network-config.yaml"),
        DATA_FOLDER.joinpath("network-config.yaml"),
    )



from typing import Callable

import importlib
import pkgutil
import click

from .accounts import AccountControllerAPI


CLI_GROUP_REGISTRATION_FN = "ape_get_cli_group"
RegisterCliEndpoint = Callable[[], click.Group]

ACCOUNTS_REGISTRATION_FN = "ape_get_accounts"
RegisterAccounts = Callable[[], AccountControllerAPI]

discovered_plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg in pkgutil.iter_modules()
    if name.startswith("ape_")
}



from collections import OrderedDict

import click

from ape.plugins import discovered_plugins, CLI_GROUP_REGISTRATION_FN


def clean_plugin_name(name: str) -> str:
    return name.replace("ape_", "").replace("_", "-")


CLI_PLUGINS = OrderedDict(
    (clean_plugin_name(n), getattr(p, CLI_GROUP_REGISTRATION_FN))
    for n, p in discovered_plugins.items()
    if hasattr(p, CLI_GROUP_REGISTRATION_FN)
)


class ApeCLI(click.MultiCommand):
    def list_commands(self, ctx):
        return list(sorted(n for n in CLI_PLUGINS))

    def get_command(self, ctx, name):
        return CLI_PLUGINS[name]()


@click.command(cls=ApeCLI, context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(message="%(version)s")
def cli():
    pass
